# ============LICENSE_START=======================================================
# Copyright (c) 2017-2018 AT&T Intellectual Property. All rights reserved.
# ================================================================================
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============LICENSE_END=========================================================
#
# ECOMP is a trademark and service mark of AT&T Intellectual Property.

"""mock of cloudify context with relationships and type_hierrarchy"""

from cloudify.mocks import MockCloudifyContext, MockNodeInstanceContext, MockNodeContext

TARGET_NODE_ID = "target_node_id"
TARGET_NODE_NAME = "target_node_name"

class MockContextNode(MockNodeContext):
    """ctx.node with type and type_hierarchy"""

    def __init__(self, id=None, properties=None, node_type=None, type_hierarchy=None):
        super(MockContextNode, self).__init__(id, properties or {})
        self._type = node_type
        self._type_hierarchy = type_hierarchy or [self._type]
        MockCloudifyContextFull.nodes[id] = self

    @property
    def type(self):
        """node type"""
        return self._type

    @property
    def type_hierarchy(self):
        """node type hierarchy is a list of types"""
        return self._type_hierarchy

class MockContextNodeInstance(MockNodeInstanceContext):
    """ctx.instance with relationships"""

    def __init__(self, id=None, runtime_properties=None, relationships=None):
        super(MockContextNodeInstance, self).__init__(id, runtime_properties or {})
        self._relationships = []
        self.add_relationships(relationships)
        MockCloudifyContextFull.instances[id] = self

    def add_relationships(self, relationships):
        """add more relationships to the node instance"""
        if not relationships:
            return
        if not self._relationships:
            self._relationships = []
        self._relationships.extend([
            MockContextRelationship(relationship)
            for relationship in (relationships or []) if TARGET_NODE_ID in relationship
        ])

    @property
    def relationships(self):
        """list of relationships to other node instances"""
        return self._relationships

class MockContextRelationshipTarget(object):
    """target of relationship"""
    def __init__(self, relationship):
        target_node_name = relationship[TARGET_NODE_NAME]
        target_node_id = relationship[TARGET_NODE_ID]

        self.node = MockCloudifyContextFull.nodes.get(target_node_name)
        self.instance = MockCloudifyContextFull.instances.get(target_node_id)

        if not self.node:
            self.node = MockContextNode(target_node_name)
        if not self.instance:
            self.instance = MockContextNodeInstance(target_node_id)

class MockContextRelationship(object):
    """item of ctx.instance.relationships"""

    def __init__(self, relationship):
        self.target = MockContextRelationshipTarget(relationship)
        self.type = relationship.get("type", "cloudify.relationships.depends_on")
        self.type_hierarchy = relationship.get("type_hierarchy") or [self.type]

class MockCloudifyContextFull(MockCloudifyContext):
    """
    ctx1 = MockCloudifyContextFull(node_id='node_1',
                                   node_name='my_1', properties={'foo': 'bar'})
    ctx2 = MockCloudifyContextFull(node_id='node_2',
                                   node_name='my_2',
                                   relationships=[{'target_node_id': 'node_1',
                                                   'target_node_name': 'my_1'}])
    """
    nodes = {}
    instances = {}

    def __init__(self,
                 node_id=None,
                 node_name=None,
                 blueprint_id=None,
                 deployment_id=None,
                 execution_id=None,
                 properties=None, node_type=None, type_hierarchy=None,
                 runtime_properties=None,
                 capabilities=None,
                 related=None,
                 source=None,
                 target=None,
                 operation=None,
                 resources=None,
                 provider_context=None,
                 bootstrap_context=None,
                 relationships=None):
        super(MockCloudifyContextFull, self).__init__(
            node_id=node_id,
            node_name=node_name,
            blueprint_id=blueprint_id,
            deployment_id=deployment_id,
            execution_id=execution_id,
            properties=properties,
            capabilities=capabilities,
            related=related,
            source=source,
            target=target,
            operation=operation,
            resources=resources,
            provider_context=provider_context,
            bootstrap_context=bootstrap_context,
            runtime_properties=runtime_properties
        )
        self._node = MockContextNode(node_name, properties, node_type, type_hierarchy)
        self._instance = MockContextNodeInstance(node_id, runtime_properties, relationships)

    @staticmethod
    def clear():
        """clean up the context links"""
        MockCloudifyContextFull.instances.clear()
        MockCloudifyContextFull.nodes.clear()



# ============LICENSE_START=======================================================
# Copyright (c) 2017-2019 AT&T Intellectual Property. All rights reserved.
# ================================================================================
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============LICENSE_END=========================================================
#
# ECOMP is a trademark and service mark of AT&T Intellectual Property.

"""tests of decorators around the cloudify operations to handle policy actions"""

import copy
import json
import logging
from datetime import datetime, timedelta
from functools import wraps

import pytest
import requests
from cloudify import ctx
from cloudify.exceptions import NonRecoverableError
from cloudify.state import current_ctx

from onap_dcae_dcaepolicy_lib import dcae_policy
from onap_dcae_dcaepolicy_lib.dcae_policy import Policies
from onap_dcae_dcaepolicy_lib.policies_output import PoliciesOutput
from tests.log_ctx import CtxLogger
from tests.mock_cloudify_ctx import (TARGET_NODE_ID, TARGET_NODE_NAME,
                                     MockCloudifyContextFull)

POLICY_ID = 'policy_id'
POLICY_VERSION = "policyVersion"
POLICY_NAME = "policyName"
POLICY_BODY = 'policy_body'
POLICY_CONFIG = 'config'
CONFIG_NAME = "ConfigName"

MONKEYED_POLICY_ID = 'monkeyed.Config_peach'
MONKEYED_POLICY_ID_2 = 'monkeyed.Config_peach_2'
MONKEYED_POLICY_ID_M = 'monkeyed.Config_multi'
MONKEYED_POLICY_ID_M_2 = 'monkeyed.Config_multi_2'
MONKEYED_POLICY_ID_M_3 = 'monkeyed.Config_multi_3'
MONKEYED_POLICY_ID_B = 'monkeyed.Config_both'
APPLICATION_CONFIG = "application_config"
LOG_FILE = 'logs/test_onap_dcae_dcaepolicy_lib.log'
LOREM_IPSUM = """Lorem ipsum dolor sit amet consectetur ametist""".split()

RUN_TS = datetime.utcnow()

class MonkeyedLogHandler(object):
    """keep the shared logger handler here"""
    _log_handler = None

    @staticmethod
    def add_handler_to(logger):
        """adds the local handler to the logger"""
        if not MonkeyedLogHandler._log_handler:
            MonkeyedLogHandler._log_handler = logging.FileHandler(LOG_FILE)
            MonkeyedLogHandler._log_handler.setLevel(logging.DEBUG)
            formatter = logging.Formatter(
                fmt='%(asctime)s.%(msecs)03d %(levelname)+8s ' + \
                    '%(threadName)s %(name)s.%(funcName)s: %(message)s', \
                datefmt='%Y%m%d_%H%M%S')
            MonkeyedLogHandler._log_handler.setFormatter(formatter)
        logger.addHandler(MonkeyedLogHandler._log_handler)

class MonkeyedPolicyBody(object):
    """policy body that policy-engine returns"""
    @staticmethod
    def create_policy_body(policy_id, policy_version=1, priority=None, **kwargs):
        """returns a fake policy-body"""
        prev_ver = policy_version - 1
        timestamp = RUN_TS + timedelta(hours=prev_ver)

        this_ver = str(policy_version)
        config = {
            "policy_updated_from_ver": str(prev_ver),
            "policy_updated_to_ver": this_ver,
            "policy_hello": LOREM_IPSUM[prev_ver % len(LOREM_IPSUM)],
            "policy_updated_ts": timestamp.isoformat()[:-3] + 'Z',
            "updated_policy_id": policy_id
        }
        config.update(copy.deepcopy(kwargs) or {})

        matching_conditions = {
            "ONAPName": "DCAE",
            CONFIG_NAME: "alex_config_name"
        }
        if priority is not None:
            matching_conditions["priority"] = priority

        return {
            "policyConfigMessage": "Config Retrieved! ",
            "policyConfigStatus": "CONFIG_RETRIEVED",
            "type": "JSON",
            POLICY_NAME: "{0}.{1}.xml".format(policy_id, this_ver),
            POLICY_VERSION: this_ver,
            POLICY_CONFIG: config,
            "matchingConditions": matching_conditions,
            "responseAttributes": {},
            "property": None
        }

    @staticmethod
    def create_policy_bare(policy_id, policy_version=1, priority=None, **kwargs):
        """returns the whole policy object for policy_id and policy_version"""
        return {
            POLICY_ID: policy_id,
            POLICY_BODY: MonkeyedPolicyBody.create_policy_body(
                policy_id, policy_version, priority, **kwargs)
        }

    @staticmethod
    def create_policy(policy_id, policy_version=1, policy_persistent=True, priority=None, **kwargs):
        """returns the whole policy object for policy_id and policy_version"""
        policy = MonkeyedPolicyBody.create_policy_bare(
            policy_id, policy_version, priority, **kwargs)
        policy[dcae_policy.POLICY_PERSISTENT] = bool(policy_persistent)
        return policy

    @staticmethod
    def is_the_same_dict(policy_body_1, policy_body_2):
        """check whether both policy_body objects are the same"""
        if not isinstance(policy_body_1, dict) or not isinstance(policy_body_2, dict):
            return False
        for key in policy_body_1:
            if key not in policy_body_2:
                return False

            val_1 = policy_body_1[key]
            val_2 = policy_body_2[key]
            if (val_1 is None) != (val_2 is None):
                return False
            if isinstance(val_1, dict) \
            and not MonkeyedPolicyBody.is_the_same_dict(val_1, val_2):
                return False
            if val_1 != val_2:
                return False
        return True

class MonkeyedNode(object):
    """node in cloudify"""
    BLUEPRINT_ID = 'test_dcae_policy_bp_id'
    DEPLOYMENT_ID = 'test_dcae_policy_dpl_id'
    EXECUTION_ID = 'test_dcae_policy_exe_id'

    def __init__(self, node_id, node_name, node_type, properties=None,
                 relationships=None, runtime_properties=None):
        self.node_id = node_id
        self.node_name = node_name
        self.ctx = MockCloudifyContextFull(
            node_id=self.node_id,
            node_name=self.node_name,
            node_type=node_type,
            blueprint_id=MonkeyedNode.BLUEPRINT_ID,
            deployment_id=MonkeyedNode.DEPLOYMENT_ID,
            execution_id=MonkeyedNode.EXECUTION_ID,
            properties=properties,
            relationships=relationships,
            runtime_properties=runtime_properties
        )

class MonkeyedResponse(object):
    """Monkey response"""
    def __init__(self, full_path, headers=None, resp_json=None):
        self.full_path = full_path
        self.status_code = 200
        self.headers = headers or {}
        self.resp_json = resp_json
        self.text = json.dumps(resp_json or {})

    def json(self):
        """returns json of response"""
        return self.resp_json

    def raise_for_status(self):
        """always happy"""
        pass

def get_app_config():
    """just get the config"""
    config = copy.deepcopy(dict(ctx.instance.runtime_properties.get(APPLICATION_CONFIG, {})))
    if not config:
        config = copy.deepcopy(dict(ctx.node.properties.get(APPLICATION_CONFIG, {})))
    return config

def operation_node_configure(**kwargs):
    """do the node-configure operation"""
    ctx.logger.info("operation_node_configure kwargs: {0}".format(kwargs))

    app_config = get_app_config()
    ctx.instance.runtime_properties[APPLICATION_CONFIG] = app_config
    ctx.instance.runtime_properties[PoliciesOutput.SERVICE_COMPONENT_NAME] = "unit_test_scn"
    ctx.logger.info("property app_config: {0}".format(json.dumps(app_config)))

@CtxLogger.log_ctx(pre_log=True, after_log=True, exe_task='exe_task')
@Policies.gather_policies_to_node()
def node_configure(**kwargs):
    """decorate with @Policies.gather_policies_to_node on policy consumer node to
    bring all policies to runtime_properties["policies"]
    """
    operation_node_configure(**kwargs)

@CtxLogger.log_ctx(pre_log=True, after_log=True, exe_task='execute_operation')
@Policies.update_policies_on_node()
def policy_update(updated_policies, removed_policies=None, **kwargs):
    """decorate with @Policies.update_policies_on_node() to update runtime_properties["policies"]

    :updated_policies: contains the list of changed policy-configs when configs_only=True (default).
    """
    # This is how to merge the policies into default app_config object
    #               (runtime_properties[APPLICATION_CONFIG] = application_config)
    app_config = get_app_config()
    ctx.logger.info("app_config {0}".format(json.dumps(app_config)))

    ctx.instance.runtime_properties[APPLICATION_CONFIG] = app_config

@CtxLogger.log_ctx(pre_log=True, after_log=True, exe_task='execute_operation')
@Policies.update_policies_on_node()
def policy_update_not_only_config(updated_policies, removed_policies=None, **kwargs):
    """decorate with @Policies.update_policies_on_node() to update runtime_properties["policies"]

    :updated_policies: contains the list of changed policy-configs when configs_only=True (default).
    """
    # This is how to merge the policies into default app_config object
    #               (runtime_properties[APPLICATION_CONFIG] = application_config)
    app_config = get_app_config()
    ctx.logger.info("app_config {0}".format(json.dumps(app_config)))

    ctx.instance.runtime_properties[APPLICATION_CONFIG] = app_config

@CtxLogger.log_ctx(pre_log=True, after_log=True, exe_task='execute_operation')
@Policies.update_policies_on_node()
def policy_update_many_calcs(updated_policies, removed_policies=None, policies=None, **kwargs):
    """decorate with @Policies.update_policies_on_node() to update runtime_properties["policies"]

    :updated_policies: contains the list of changed policy-configs when configs_only=True (default).
    """
    app_config = get_app_config()
    ctx.logger.info("app_config {0}".format(json.dumps(app_config)))

    ctx.instance.runtime_properties[APPLICATION_CONFIG] = app_config


@CtxLogger.log_ctx(pre_log=True, after_log=True, exe_task='exe_task')
@Policies.cleanup_policies_on_node
def node_delete(**kwargs):
    """delete <service_component_name> records in consul-kv"""
    operation_node_configure(**kwargs)


class CurrentCtx(object):
    """cloudify context"""
    _node_ms = None

    @staticmethod
    def set_current_ctx(include_bad=True, include_good=True):
        """set up the nodes for cloudify"""
        Policies._init()

        def add_target_to_relationships(relationships, node):
            """adds the node as the target of relationships"""
            relationships.append({TARGET_NODE_ID: node.node_id, TARGET_NODE_NAME: node.node_name})

        relationships = []
        if include_good:
            node_policy = MonkeyedNode(
                'dcae_policy_node_id',
                'dcae_policy_node_name',
                dcae_policy.DCAE_POLICY_TYPE,
                {POLICY_ID: MONKEYED_POLICY_ID},
                None,
                {POLICY_BODY: MonkeyedPolicyBody.create_policy_body(
                    MONKEYED_POLICY_ID, priority="1")}
            )
            add_target_to_relationships(relationships, node_policy)

        if include_bad:
            bad_policy_2 = MonkeyedNode(
                'bad_policy_2_node_id',
                'bad_policy_2_node_name',
                dcae_policy.DCAE_POLICY_TYPE,
                {POLICY_ID: MONKEYED_POLICY_ID_2},
                None,
                None
            )
            add_target_to_relationships(relationships, bad_policy_2)

        if include_good:
            node_policy_2 = MonkeyedNode(
                'dcae_policy_node_id_2',
                'dcae_policy_node_name_2',
                dcae_policy.DCAE_POLICY_TYPE,
                {POLICY_ID: MONKEYED_POLICY_ID_2},
                None,
                {POLICY_BODY: MonkeyedPolicyBody.create_policy_body(
                    MONKEYED_POLICY_ID_2, 4, priority="2")}
            )
            add_target_to_relationships(relationships, node_policy_2)

        if include_bad:
            bad_policy_3 = MonkeyedNode(
                'bad_policy_3_node_id',
                'bad_policy_3_node_name',
                dcae_policy.DCAE_POLICY_TYPE,
                {POLICY_ID: MONKEYED_POLICY_ID_2},
                None,
                None
            )
            add_target_to_relationships(relationships, bad_policy_3)

            bad_policy_4 = MonkeyedNode(
                'bad_policy_4_node_id',
                'bad_policy_4_node_name',
                dcae_policy.DCAE_POLICY_TYPE,
                None,
                None,
                None
            )
            add_target_to_relationships(relationships, bad_policy_4)

            weird_policy_5 = MonkeyedNode(
                'weird_policy_5_node_id',
                'weird_policy_5_node_name',
                dcae_policy.DCAE_POLICY_TYPE,
                None,
                None,
                {POLICY_BODY: MonkeyedPolicyBody.create_policy_body(
                    MONKEYED_POLICY_ID, 3, priority="2")}
            )
            add_target_to_relationships(relationships, weird_policy_5)

        if include_good:
            node_policies = MonkeyedNode(
                'dcae_policies_node_id',
                'dcae_policies_node_name',
                dcae_policy.DCAE_POLICIES_TYPE,
                {dcae_policy.POLICY_FILTER: {
                    POLICY_NAME: MONKEYED_POLICY_ID_M + ".*",
                    dcae_policy.CONFIG_ATTRIBUTES: json.dumps({
                        CONFIG_NAME: "alex_config_name"
                    })
                }},
                None,
                {dcae_policy.POLICIES_FILTERED: {
                    MONKEYED_POLICY_ID_M:
                        MonkeyedPolicyBody.create_policy_bare(MONKEYED_POLICY_ID_M, 3)}}
            )
            add_target_to_relationships(relationships, node_policies)

            node_policies_2 = MonkeyedNode(
                'dcae_policies_2_node_id',
                'dcae_policies_2_node_name',
                dcae_policy.DCAE_POLICIES_TYPE,
                {dcae_policy.POLICY_FILTER: {POLICY_NAME: MONKEYED_POLICY_ID_M + ".*"}},
                None,
                {dcae_policy.POLICIES_FILTERED: {
                    MONKEYED_POLICY_ID_M:
                        MonkeyedPolicyBody.create_policy_bare(MONKEYED_POLICY_ID_M, 2)}}
            )
            add_target_to_relationships(relationships, node_policies_2)

            policies_empty = MonkeyedNode(
                'dcae_policies_empty_node_id',
                'dcae_policies_empty_node_name',
                dcae_policy.DCAE_POLICIES_TYPE,
                {dcae_policy.POLICY_FILTER: {"empty": None}},
                None,
                None
            )
            add_target_to_relationships(relationships, policies_empty)

            policies_empty_2 = MonkeyedNode(
                'dcae_policies_empty_2_node_id',
                'dcae_policies_empty_2_node_name',
                dcae_policy.DCAE_POLICIES_TYPE,
                None,
                None,
                None
            )
            add_target_to_relationships(relationships, policies_empty_2)

        non_policies = MonkeyedNode(
            'non_policies_node_id',
            'non_policies_node_name',
            "non.policy.type"
        )
        add_target_to_relationships(relationships, non_policies)

        if include_good:
            node_policies_b = MonkeyedNode(
                'dcae_policies_b_node_id',
                'dcae_policies_b_node_name',
                dcae_policy.DCAE_POLICIES_TYPE,
                {dcae_policy.POLICY_FILTER: {POLICY_NAME: MONKEYED_POLICY_ID_M + ".*"}},
                None,
                {dcae_policy.POLICIES_FILTERED: {
                    MONKEYED_POLICY_ID_B:
                        MonkeyedPolicyBody.create_policy_bare(MONKEYED_POLICY_ID_B, 1)}}
            )
            add_target_to_relationships(relationships, node_policies_b)

            node_policies_b_2 = MonkeyedNode(
                'dcae_policies_b_2_node_id',
                'dcae_policies_b_2_node_name',
                dcae_policy.DCAE_POLICIES_TYPE,
                {dcae_policy.POLICY_FILTER: {POLICY_NAME: MONKEYED_POLICY_ID_M + ".*"}},
                None,
                {dcae_policy.POLICIES_FILTERED: {
                    MONKEYED_POLICY_ID_B:
                        MonkeyedPolicyBody.create_policy_bare(MONKEYED_POLICY_ID_B, 2)}}
            )
            add_target_to_relationships(relationships, node_policies_b_2)

        if include_good:
            node_policy_b = MonkeyedNode(
                'dcae_policy_b_node_id',
                'dcae_policy_b_node_name',
                dcae_policy.DCAE_POLICY_TYPE,
                {POLICY_ID: MONKEYED_POLICY_ID_B},
                None,
                {POLICY_BODY: MonkeyedPolicyBody.create_policy_body(
                    MONKEYED_POLICY_ID_B, 4, priority="1.5")}
            )
            add_target_to_relationships(relationships, node_policy_b)

            node_policies_b_5 = MonkeyedNode(
                'dcae_policies_b_5_node_id',
                'dcae_policies_b_5_node_name',
                dcae_policy.DCAE_POLICIES_TYPE,
                {dcae_policy.POLICY_FILTER: {POLICY_NAME: MONKEYED_POLICY_ID_M + ".*"}},
                None,
                {dcae_policy.POLICIES_FILTERED: {
                    MONKEYED_POLICY_ID_B:
                        MonkeyedPolicyBody.create_policy_bare(MONKEYED_POLICY_ID_B, 5)}}
            )
            add_target_to_relationships(relationships, node_policies_b_5)


        CurrentCtx._node_ms = MonkeyedNode(
            'test_ms_id', 'test_ms_name', "ms.nodes.type",
            {APPLICATION_CONFIG: MonkeyedPolicyBody.create_policy_body(
                 "no_policy", db_port="123", weather="snow")[POLICY_CONFIG]
            },
            relationships
        )
        current_ctx.set(CurrentCtx._node_ms.ctx)
        MonkeyedLogHandler.add_handler_to(CurrentCtx._node_ms.ctx.logger)

    @staticmethod
    def reset():
        """reset context"""
        current_ctx.set(CurrentCtx._node_ms.ctx)


def monkeyed_consul_boom(full_path, json):
    """boom on monkeypatch for the put to consul"""
    raise requests.ConnectionError("monkey-boom")


@pytest.fixture()
def fix_consul_boom(monkeypatch):
    """monkeyed discovery request.put"""
    PoliciesOutput._lazy_inited = False
    monkeypatch.setattr('requests.put', monkeyed_consul_boom)
    yield fix_consul_boom
    PoliciesOutput._lazy_inited = False


def monkeyed_consul_put(full_path, json):
    """monkeypatch for the put to consul"""
    return MonkeyedResponse(full_path)


@pytest.fixture()
def fix_consul(monkeypatch):
    """monkeyed discovery request.put"""
    PoliciesOutput._lazy_inited = False
    monkeypatch.setattr('requests.put', monkeyed_consul_put)
    yield fix_consul
    PoliciesOutput._lazy_inited = False


def cfy_ctx(include_bad=True, include_good=True):
    """test and safely clean up"""
    def cfy_ctx_decorator(func):
        """test and safely clean up"""
        if not func:
            return

        @wraps(func)
        def ctx_wrapper(*args, **kwargs):
            """test"""
            try:
                CurrentCtx.set_current_ctx(include_bad, include_good)

                func(*args, **kwargs)

            finally:
                ctx.logger.info("MockCloudifyContextFull.clear")
                MockCloudifyContextFull.clear()
                current_ctx.clear()

        return ctx_wrapper
    return cfy_ctx_decorator

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_gather_policies_to_node():
    """test gather_policies_to_node"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_policies_to_node():
    """test gather_policies_to_node"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    assert MONKEYED_POLICY_ID in policies
    expected_1 = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID, priority="1")
    policy = policies[MONKEYED_POLICY_ID]
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID, json.dumps(expected_1)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_1)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_1, policy)

    assert MONKEYED_POLICY_ID_B in policies
    expected_b = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_B, 4, priority="1.5")
    policy = policies[MONKEYED_POLICY_ID_B]
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID_B, json.dumps(expected_b)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_B, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_b)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_b, policy)

    assert MONKEYED_POLICY_ID_2 in policies
    expected_2 = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_2, 4, priority="2")
    policy = policies[MONKEYED_POLICY_ID_2]
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID_2, json.dumps(expected_2)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_2)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_2, policy)

    assert MONKEYED_POLICY_ID_M in policies
    expected_m = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_M, 2, False)
    policy = policies[MONKEYED_POLICY_ID_M]
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID_M, json.dumps(expected_m)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_M, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_m)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_m, policy)

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_update_policies():
    """test policy_update"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    updated_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_2, 2, priority="aa20")
    added_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_M_2, 2,
                                                    False, priority="1")
    ctx.logger.info("policy_update: [{0}]".format(json.dumps(updated_policy)))

    ctx.logger.info("policy[{0}]: not yet in policies".format(MONKEYED_POLICY_ID_M_2))
    assert MONKEYED_POLICY_ID_M_2 not in policies

    policy_filter_ids = list(runtime_properties.get(dcae_policy.POLICY_FILTERS) or ["--"])

    policy_update(updated_policies=[updated_policy],
                  added_policies={
                      policy_filter_ids[0]: {
                          dcae_policy.POLICIES: {MONKEYED_POLICY_ID_M_2: added_policy}}
                  },
                  removed_policies=[MONKEYED_POLICY_ID_M])

    ctx.logger.info("policy[{0}]: removed".format(MONKEYED_POLICY_ID_M))
    assert MONKEYED_POLICY_ID_M not in policies

    assert MONKEYED_POLICY_ID_M_2 in policies
    policy = policies[MONKEYED_POLICY_ID_M_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_M_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, added_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(added_policy, policy)

    assert MONKEYED_POLICY_ID_2 in policies
    policy = policies[MONKEYED_POLICY_ID_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, updated_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(updated_policy, policy)

    assert MONKEYED_POLICY_ID in policies
    policy = policies[MONKEYED_POLICY_ID]
    expected_1 = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID, priority="1")
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID, json.dumps(expected_1)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_1)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_1, policy)

    assert MONKEYED_POLICY_ID_B in policies
    policy = policies[MONKEYED_POLICY_ID_B]
    expected_b = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_B, 4, priority="1.5")
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID_B, json.dumps(expected_1)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_B, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_b)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_b, policy)

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_update_not_only_config():
    """test policy_update"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    updated_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_2, 2, priority="aa20")
    added_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_M_2, 2,
                                                    False, priority="1")
    ctx.logger.info("policy_update: [{0}]".format(json.dumps(updated_policy)))

    ctx.logger.info("policy[{0}]: not yet in policies".format(MONKEYED_POLICY_ID_M_2))
    assert MONKEYED_POLICY_ID_M_2 not in policies

    policy_filter_ids = list(runtime_properties.get(dcae_policy.POLICY_FILTERS) or ["--"])

    policy_update_not_only_config(updated_policies=[updated_policy],
                                  added_policies={
                                      policy_filter_ids[0]: {
                                          dcae_policy.POLICIES: {MONKEYED_POLICY_ID_M_2: added_policy}}
                                  },
                                  removed_policies=[MONKEYED_POLICY_ID_M])

    ctx.logger.info("policy[{0}]: removed".format(MONKEYED_POLICY_ID_M))
    assert MONKEYED_POLICY_ID_M not in policies

    assert MONKEYED_POLICY_ID_M_2 in policies
    policy = policies[MONKEYED_POLICY_ID_M_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_M_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, added_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(added_policy, policy)

    assert MONKEYED_POLICY_ID_2 in policies
    policy = policies[MONKEYED_POLICY_ID_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, updated_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(updated_policy, policy)

    assert MONKEYED_POLICY_ID in policies
    policy = policies[MONKEYED_POLICY_ID]
    expected_1 = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID, priority="1")
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID, json.dumps(expected_1)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_1)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_1, policy)

    assert MONKEYED_POLICY_ID_B in policies
    policy = policies[MONKEYED_POLICY_ID_B]
    expected_b = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_B, 4, priority="1.5")
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID_B, json.dumps(expected_1)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_B, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_b)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_b, policy)

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_update_policies_not():
    """test policy_update - ignore all policies with junk params"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    assert APPLICATION_CONFIG in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    app_config = runtime_properties[APPLICATION_CONFIG]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))
    ctx.logger.info("app_config: {0}".format(json.dumps(app_config)))

    expected_policies = copy.deepcopy(policies)
    expected_app_config = copy.deepcopy(app_config)

    existing_policy = MonkeyedPolicyBody.create_policy_bare(MONKEYED_POLICY_ID, priority="1")
    damaged_policy = MonkeyedPolicyBody.create_policy_bare(MONKEYED_POLICY_ID_2)
    del damaged_policy[POLICY_BODY][POLICY_CONFIG]
    updated_policy = MonkeyedPolicyBody.create_policy_bare(MONKEYED_POLICY_ID_M_3, 3)
    added_policy = MonkeyedPolicyBody.create_policy_bare(MONKEYED_POLICY_ID_M_2, 4)
    junk_policy_filter_id = "<<<junk_removed_policy_id>>>"
    unexpected_removed_policy_id = "<<<junk_removed_policy_id>>>"
    ctx.logger.info("policy_update: [{0}]".format(json.dumps(updated_policy)))

    assert MONKEYED_POLICY_ID in policies
    assert MONKEYED_POLICY_ID_2 in policies
    assert MONKEYED_POLICY_ID_M_2 not in policies
    assert MONKEYED_POLICY_ID_M_3 not in policies
    assert unexpected_removed_policy_id not in policies
    assert junk_policy_filter_id not in runtime_properties.get(dcae_policy.POLICY_FILTERS, {})

    policy_filter_ids = list(runtime_properties.get(dcae_policy.POLICY_FILTERS) or ["--"])

    policy_update(updated_policies=[existing_policy, damaged_policy, updated_policy],
                  added_policies={
                      junk_policy_filter_id: {
                          dcae_policy.POLICIES: {MONKEYED_POLICY_ID_M_2: added_policy}},
                      policy_filter_ids[0]: {
                          dcae_policy.POLICIES: {MONKEYED_POLICY_ID_2: damaged_policy}}
                  },
                  removed_policies=[unexpected_removed_policy_id])

    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    ctx.logger.info("policies not changed: {0}".format(json.dumps(policies)))
    assert MonkeyedPolicyBody.is_the_same_dict(policies, expected_policies)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_policies, policies)

    ctx.logger.info("app_config not changed: {0}".format(json.dumps(app_config)))
    assert MonkeyedPolicyBody.is_the_same_dict(app_config, expected_app_config)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_app_config, app_config)

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_update_many_calcs():
    """test policy_update"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    updated_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_2, 2, priority="aa20")
    added_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_M_2, 2,
                                                    False, priority="1")
    ctx.logger.info("policy_update: [{0}]".format(json.dumps(updated_policy)))

    ctx.logger.info("policy[{0}]: not yet in policies".format(MONKEYED_POLICY_ID_M_2))
    assert MONKEYED_POLICY_ID_M_2 not in policies

    policy_filter_ids = list(runtime_properties.get(dcae_policy.POLICY_FILTERS) or ["--"])

    policy_update_many_calcs(updated_policies=[updated_policy],
                             added_policies={
                                 policy_filter_ids[0]: {
                                     dcae_policy.POLICIES: {MONKEYED_POLICY_ID_M_2: added_policy}}
                             },
                             removed_policies=[MONKEYED_POLICY_ID_M])

    ctx.logger.info("policy[{0}]: removed".format(MONKEYED_POLICY_ID_M))
    assert MONKEYED_POLICY_ID_M not in policies

    assert MONKEYED_POLICY_ID_M_2 in policies
    policy = policies[MONKEYED_POLICY_ID_M_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_M_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, added_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(added_policy, policy)

    assert MONKEYED_POLICY_ID_2 in policies
    policy = policies[MONKEYED_POLICY_ID_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, updated_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(updated_policy, policy)

    assert MONKEYED_POLICY_ID in policies
    policy = policies[MONKEYED_POLICY_ID]
    expected_1 = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID, priority="1")
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID, json.dumps(expected_1)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_1)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_1, policy)

    assert MONKEYED_POLICY_ID_B in policies
    policy = policies[MONKEYED_POLICY_ID_B]
    expected_b = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_B, 4, priority="1.5")
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID_B, json.dumps(expected_1)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_B, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_b)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_b, policy)

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_remove_all_policies():
    """test policy_update - remove all policies"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    remove_policy_ids = list(policies)

    policy_update(updated_policies=None, added_policies=None, removed_policies=remove_policy_ids)

    ctx.logger.info("removed: {0}".format(remove_policy_ids))
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))
    assert dcae_policy.POLICIES in runtime_properties
    assert Policies.get_policy_bodies() == []

    assert APPLICATION_CONFIG in runtime_properties
    assert APPLICATION_CONFIG in ctx.node.properties
    app_config = runtime_properties[APPLICATION_CONFIG]
    expected_config = dict(ctx.node.properties[APPLICATION_CONFIG])
    ctx.logger.info("expected = default application_config: {0}".format(json.dumps(app_config)))
    assert MonkeyedPolicyBody.is_the_same_dict(app_config, expected_config)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_config, app_config)

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_remove_all_policies_twice():
    """test policy_update - remove all policies twice"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    remove_policy_ids = list(policies)

    policy_update(updated_policies=None, added_policies=None, removed_policies=remove_policy_ids)
    policy_update(updated_policies=None, added_policies=None, removed_policies=remove_policy_ids)

    ctx.logger.info("removed: {0}".format(remove_policy_ids))
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))
    assert dcae_policy.POLICIES in runtime_properties
    assert Policies.get_policy_bodies() == []

    assert APPLICATION_CONFIG in runtime_properties
    assert APPLICATION_CONFIG in ctx.node.properties
    app_config = runtime_properties[APPLICATION_CONFIG]
    expected_config = dict(ctx.node.properties[APPLICATION_CONFIG])
    ctx.logger.info("expected = default application_config: {0}".format(json.dumps(app_config)))
    assert MonkeyedPolicyBody.is_the_same_dict(app_config, expected_config)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_config, app_config)

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_remove_then_update():
    """test policy_update"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    remove_policy_ids = list(policies)
    policy_update(updated_policies=None, added_policies=None, removed_policies=remove_policy_ids)

    updated_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_2, 2, priority="aa20")
    added_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_M_2, 2,
                                                    False, priority="1")
    ctx.logger.info("policy_update: [{0}]".format(json.dumps(updated_policy)))

    ctx.logger.info("policy[{0}]: not yet in policies".format(MONKEYED_POLICY_ID_M_2))
    assert MONKEYED_POLICY_ID_M_2 not in policies

    policy_filter_ids = list(runtime_properties.get(dcae_policy.POLICY_FILTERS) or ["--"])

    policy_update(updated_policies=[updated_policy],
                  added_policies={
                      policy_filter_ids[0]: {
                          dcae_policy.POLICIES: {MONKEYED_POLICY_ID_M_2: added_policy}}
                  },
                  removed_policies=[MONKEYED_POLICY_ID_M])

    ctx.logger.info("policy[{0}]: removed".format(MONKEYED_POLICY_ID_M))
    assert MONKEYED_POLICY_ID_M not in policies

    assert MONKEYED_POLICY_ID_M_2 in policies
    policy = policies[MONKEYED_POLICY_ID_M_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_M_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, added_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(added_policy, policy)

    assert MONKEYED_POLICY_ID_2 in policies
    policy = policies[MONKEYED_POLICY_ID_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, updated_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(updated_policy, policy)

    assert MONKEYED_POLICY_ID in policies
    assert MONKEYED_POLICY_ID_B in policies

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_remove_update_many_calcs():
    """test policy_update"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    remove_policy_ids = list(policies)
    policy_update_many_calcs(updated_policies=None,
                             added_policies=None,
                             removed_policies=remove_policy_ids)

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    updated_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_2, 2, priority="aa20")
    added_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_M_2, 2,
                                                    False, priority="1")
    ctx.logger.info("policy_update: [{0}]".format(json.dumps(updated_policy)))

    ctx.logger.info("policy[{0}]: not yet in policies".format(MONKEYED_POLICY_ID_M_2))
    assert MONKEYED_POLICY_ID_M_2 not in policies

    policy_filter_ids = list(runtime_properties.get(dcae_policy.POLICY_FILTERS) or ["--"])

    policy_update_many_calcs(updated_policies=[updated_policy],
                             added_policies={
                                 policy_filter_ids[0]: {
                                     dcae_policy.POLICIES: {MONKEYED_POLICY_ID_M_2: added_policy}}
                             },
                             removed_policies=[MONKEYED_POLICY_ID_M])

    ctx.logger.info("policy[{0}]: removed".format(MONKEYED_POLICY_ID_M))
    assert MONKEYED_POLICY_ID_M not in policies

    assert MONKEYED_POLICY_ID_M_2 in policies
    policy = policies[MONKEYED_POLICY_ID_M_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_M_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, added_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(added_policy, policy)

    assert MONKEYED_POLICY_ID_2 in policies
    policy = policies[MONKEYED_POLICY_ID_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, updated_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(updated_policy, policy)

    assert MONKEYED_POLICY_ID in policies
    assert MONKEYED_POLICY_ID_B in policies

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_bad_update_many_calcs():
    """test policy_update"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    damaged_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_2, 2, priority="aa20")
    damaged_policy[POLICY_BODY][POLICY_CONFIG] = ["damaged config"]

    added_policy = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_M_2, 2,
                                                    False, priority="1")
    added_policy[POLICY_BODY][POLICY_CONFIG] = {"unexpected":"foo", "none": None}

    ctx.logger.info("policy_update: [{0}]".format(json.dumps(damaged_policy)))

    ctx.logger.info("policy[{0}]: not yet in policies".format(MONKEYED_POLICY_ID_M_2))
    assert MONKEYED_POLICY_ID_M_2 not in policies

    policy_filter_ids = list(runtime_properties.get(dcae_policy.POLICY_FILTERS) or ["--"])

    policy_update_many_calcs(updated_policies=[damaged_policy],
                             added_policies={
                                 policy_filter_ids[0]: {
                                     dcae_policy.POLICIES: {MONKEYED_POLICY_ID_M_2: added_policy}}
                             },
                             removed_policies=[MONKEYED_POLICY_ID_M])

    ctx.logger.info("policy[{0}]: removed".format(MONKEYED_POLICY_ID_M))
    assert MONKEYED_POLICY_ID_M not in policies

    assert MONKEYED_POLICY_ID_M_2 in policies
    policy = policies[MONKEYED_POLICY_ID_M_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_M_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, added_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(added_policy, policy)

    assert MONKEYED_POLICY_ID_2 in policies
    policy = policies[MONKEYED_POLICY_ID_2]
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_2, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, damaged_policy)
    assert MonkeyedPolicyBody.is_the_same_dict(damaged_policy, policy)

    assert MONKEYED_POLICY_ID in policies
    policy = policies[MONKEYED_POLICY_ID]
    expected_1 = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID, priority="1")
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID, json.dumps(expected_1)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_1)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_1, policy)

    assert MONKEYED_POLICY_ID_B in policies
    policy = policies[MONKEYED_POLICY_ID_B]
    expected_b = MonkeyedPolicyBody.create_policy(MONKEYED_POLICY_ID_B, 4, priority="1.5")
    ctx.logger.info("expected[{0}]: {1}".format(MONKEYED_POLICY_ID_B, json.dumps(expected_1)))
    ctx.logger.info("policy[{0}]: {1}".format(MONKEYED_POLICY_ID_B, json.dumps(policy)))
    assert MonkeyedPolicyBody.is_the_same_dict(policy, expected_b)
    assert MonkeyedPolicyBody.is_the_same_dict(expected_b, policy)

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True, include_good=False)
def test_bad_policies():
    """test bad policy nodes"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True, include_good=False)
def test_wrong_ctx_node_configure():
    """test wrong ctx"""
    current_ctx.set(ctx.instance.relationships[0])
    ctx_type = ctx.type

    with pytest.raises(NonRecoverableError) as excinfo:
        node_configure()

    CurrentCtx.reset()
    ctx.logger.info("{0} not a node boom: {1}".format(ctx_type, str(excinfo.value)))
    assert ctx_type == 'cloudify.relationships.depends_on'
    assert str(excinfo.value) == "can only invoke gather_policies_to_node on node"

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True, include_good=False)
def test_wrong_ctx_policy_update():
    """test wrong ctx"""
    no_policies = Policies.get_policy_bodies()
    current_ctx.set(ctx.instance.relationships[0])
    ctx_type = ctx.type

    with pytest.raises(NonRecoverableError) as excinfo:
        policy_update(updated_policies=None, added_policies=None, removed_policies=None)

    CurrentCtx.reset()
    ctx.logger.info("{0} not a node boom: {1}".format(ctx_type, str(excinfo.value)))
    assert ctx_type == 'cloudify.relationships.depends_on'
    assert no_policies == []
    assert str(excinfo.value) == "can only invoke update_policies_on_node on node"

def test_defenses_on_decorators():
    """test defenses of code mainly for code coverage"""
    should_be_none = Policies.gather_policies_to_node()(None)
    assert should_be_none is None

    should_be_none = Policies.update_policies_on_node()(None)
    assert should_be_none is None

def monkeyed_set_policies_boom(policies):
    """monkeypatch for exception"""
    raise Exception("Policies._set_policies boom: {0}".format(json.dumps(policies or {})))

@pytest.fixture()
def fix_boom_gather(monkeypatch):
    """monkeyed boom"""
    monkeypatch.setattr('onap_dcae_dcaepolicy_lib.dcae_policy.Policies._set_policies',
                        monkeyed_set_policies_boom)
    return fix_boom_gather  # provide the fixture value

@cfy_ctx(include_bad=True, include_good=False)
@pytest.mark.usefixtures("fix_boom_gather")
def test_exception_on_gather():
    """test exception on gather"""
    with pytest.raises(NonRecoverableError) as excinfo:
        node_configure()

    ctx.logger.info("monkeyed_set_policies_boom: {0}".format(str(excinfo.value)))
    assert str(excinfo.value).startswith("Failed to set the policies ")

def monkeyed_update_policies_boom(policies):
    """monkeypatch for exception"""
    raise Exception("Policies._update_policies boom")

@pytest.fixture()
def fix_boom_update_policies(monkeypatch):
    """monkeyed boom"""
    monkeypatch.setattr('onap_dcae_dcaepolicy_lib.dcae_policy.Policies._update_policies',
                        monkeyed_update_policies_boom)
    return fix_boom_update_policies  # provide the fixture value

@cfy_ctx(include_bad=True, include_good=False)
@pytest.mark.usefixtures("fix_boom_update_policies")
def test_exception_on_update():
    """test exception on update_policies"""
    with pytest.raises(NonRecoverableError) as excinfo:
        policy_update(updated_policies=None, added_policies=None, removed_policies=None)

    ctx.logger.info("monkeyed_update_policies_boom: {0}".format(str(excinfo.value)))
    assert str(excinfo.value).startswith("Failed to update the policies ")

@cfy_ctx(include_bad=True, include_good=False)
def test_defenses_on_policy_update():
    """test defenses on policy_update"""
    policy_update(updated_policies=None, added_policies=None, removed_policies=None)
    ctx.logger.info("policy_update() ok")

@cfy_ctx(include_bad=True, include_good=False)
def test_defenses_on_set_policies():
    """test defenses of code mainly for code coverage"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    Policies._set_policies({})

    assert dcae_policy.POLICIES not in runtime_properties

    Policies._set_policies({})

    assert dcae_policy.POLICIES not in runtime_properties

@pytest.mark.usefixtures("fix_consul")
@cfy_ctx(include_bad=True)
def test_delete_node():
    """test delete"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    node_delete()


@pytest.mark.usefixtures("fix_consul_boom")
@cfy_ctx(include_bad=True)
def test_delete_node_no_consul():
    """test delete without consul"""
    node_configure()

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES in runtime_properties
    policies = runtime_properties[dcae_policy.POLICIES]
    ctx.logger.info("policies: {0}".format(json.dumps(policies)))

    node_delete()


@pytest.mark.usefixtures("fix_consul_boom")
@cfy_ctx(include_bad=True)
def test_delete_node_no_policies():
    """test delete without consul and setup"""

    ctx.instance.runtime_properties[PoliciesOutput.POLICIES_EVENT] = {}
    ctx.instance.runtime_properties[PoliciesOutput.SERVICE_COMPONENT_NAME] = "delete_node_empty"

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES not in runtime_properties

    node_delete()


@pytest.mark.usefixtures("fix_consul_boom")
@cfy_ctx(include_bad=True)
def test_delete_node_empty():
    """test delete without consul and setup"""
    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES not in runtime_properties

    node_delete()

@pytest.mark.usefixtures("fix_consul_boom")
@cfy_ctx(include_bad=True)
def test_delete_node_lost_scn():
    """test delete without consul and setup"""
    ctx.instance.runtime_properties[PoliciesOutput.POLICIES_EVENT] = {}

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES not in runtime_properties

    node_delete()


@pytest.mark.usefixtures("fix_consul_boom")
@cfy_ctx(include_bad=True)
def test_delete_node_empty_config():
    """test delete without consul and setup"""

    ctx.instance.runtime_properties[PoliciesOutput.POLICIES_EVENT] = {}
    ctx.instance.runtime_properties[PoliciesOutput.SERVICE_COMPONENT_NAME] = "delete_node_empty"

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES not in runtime_properties

    node_delete()

@pytest.mark.usefixtures("fix_consul_boom")
@cfy_ctx(include_bad=True)
def test_delete_ms_no_consul_addr():
    """test delete without consul and setup"""

    ctx.instance.runtime_properties[PoliciesOutput.POLICIES_EVENT] = {}
    ctx.instance.runtime_properties[PoliciesOutput.SERVICE_COMPONENT_NAME] = "delete_node_empty"

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES not in runtime_properties

    node_delete()


@pytest.mark.usefixtures("fix_consul_boom")
@cfy_ctx(include_bad=True)
def test_delete_bad_config():
    """test delete without consul and setup"""

    ctx.instance.runtime_properties[PoliciesOutput.POLICIES_EVENT] = {}
    ctx.instance.runtime_properties[PoliciesOutput.SERVICE_COMPONENT_NAME] = "delete_node_empty"

    runtime_properties = ctx.instance.runtime_properties
    ctx.logger.info("runtime_properties: {0}".format(json.dumps(runtime_properties)))

    assert dcae_policy.POLICIES not in runtime_properties

    node_delete()




# ================================================================================
# Copyright (c) 2017-2018 AT&T Intellectual Property. All rights reserved.
# ================================================================================
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============LICENSE_END=========================================================
#
# ECOMP is a trademark and service mark of AT&T Intellectual Property.

""":@CtxLogger.log_ctx: decorator for logging the cloudify ctx before and after operation"""

import json
import traceback
from functools import wraps

from cloudify import ctx
from cloudify.context import NODE_INSTANCE, RELATIONSHIP_INSTANCE

class CtxLogger(object):
    """static class for logging cloudify context ctx"""
    @staticmethod
    def _get_ctx_node_info(ctx_node):
        if not ctx_node:
            return {}
        return {'id': ctx_node.id, 'name': ctx_node.name, 'type': ctx_node.type,
                'type_hierarchy': ctx_node.type_hierarchy, 'properties': ctx_node.properties}

    @staticmethod
    def _get_ctx_instance_info(ctx_instance):
        if not ctx_instance:
            return {}
        return {'id' : ctx_instance.id, 'runtime_properties' : ctx_instance.runtime_properties,
                'relationships' : CtxLogger._get_ctx_instance_relationships_info(ctx_instance)}

    @staticmethod
    def _get_ctx_instance_relationships_info(ctx_instance):
        if not ctx_instance or not ctx_instance.relationships:
            return []
        return [{'target': CtxLogger._get_ctx_source_target_info(r.target), \
                 'type':r.type, 'type_hierarchy':r.type_hierarchy} \
                for r in ctx_instance.relationships]

    @staticmethod
    def _get_ctx_source_target_info(ctx_source_target):
        if not ctx_source_target:
            return {}
        return {'node': CtxLogger._get_ctx_node_info(ctx_source_target.node),
                'instance' : CtxLogger._get_ctx_instance_info(ctx_source_target.instance)}

    @staticmethod
    def get_ctx_info():
        """collect the context data from ctx"""
        context = {
            'type': ctx.type,
            'blueprint.id': ctx.blueprint.id,
            'deployment.id': ctx.deployment.id,
            'execution_id': ctx.execution_id,
            'workflow_id': ctx.workflow_id,
            'task_id': ctx.task_id,
            'task_name': ctx.task_name,
            'task_queue': ctx.task_queue,
            'task_target': ctx.task_target,
            'operation': {
                'name': ctx.operation.name,
                'retry_number': ctx.operation.retry_number,
                'max_retries': ctx.operation.max_retries
            },
            'plugin': {
                'name': ctx.plugin.name,
                'package_name': ctx.plugin.package_name,
                'package_version': ctx.plugin.package_version,
                'prefix': ctx.plugin.prefix,
                'workdir': ctx.plugin.workdir
            }
        }
        if ctx.type == NODE_INSTANCE:
            context['node'] = CtxLogger._get_ctx_node_info(ctx.node)
            context['instance'] = CtxLogger._get_ctx_instance_info(ctx.instance)
        elif ctx.type == RELATIONSHIP_INSTANCE:
            context['source'] = CtxLogger._get_ctx_source_target_info(ctx.source)
            context['target'] = CtxLogger._get_ctx_source_target_info(ctx.target)

        return context

    @staticmethod
    def log_ctx_info(func_name):
        """shortcut for logging of the ctx of the function"""
        try:
            if ctx.type == NODE_INSTANCE:
                ctx.logger.info("{0} {1} context: {2}".format(\
                    func_name, ctx.instance.id, json.dumps(CtxLogger.get_ctx_info())))
            elif ctx.type == RELATIONSHIP_INSTANCE:
                ctx.logger.info("{0} context: {1}".format(\
                    func_name, json.dumps(CtxLogger.get_ctx_info())))
        except Exception as ex:
            ctx.logger.error("Failed to log the node context: {0}: {1}" \
                .format(str(ex), traceback.format_exc()))

    @staticmethod
    def log_ctx(pre_log=True, after_log=False, exe_task=None):
        """Decorate each operation on the node to log the context - before and after.
        Optionally save the current function name into runtime_properties[exe_task]
        """
        def log_ctx_info_decorator(func, **arguments):
            """Decorate each operation on the node to log the context"""
            if func is not None:
                @wraps(func)
                def wrapper(*args, **kwargs):
                    """the actual logger before and after"""
                    try:
                        if ctx.type == NODE_INSTANCE and exe_task:
                            ctx.instance.runtime_properties[exe_task] = func.__name__
                    except Exception as ex:
                        ctx.logger.error("Failed to set exe_task {0}: {1}: {2}" \
                            .format(exe_task, str(ex), traceback.format_exc()))
                    if pre_log:
                        CtxLogger.log_ctx_info('before ' + func.__name__)

                    result = func(*args, **kwargs)

                    if after_log:
                        CtxLogger.log_ctx_info('after ' + func.__name__)

                    return result
                return wrapper
        return log_ctx_info_decorator



# ================================================================================
# Copyright (c) 2017-2019 AT&T Intellectual Property. All rights reserved.
# ================================================================================
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============LICENSE_END=========================================================
#
# ECOMP is a trademark and service mark of AT&T Intellectual Property.

"""dcae_policy contains decorators for the policy lifecycle in cloudify"""

import json
import traceback
from copy import deepcopy
from functools import wraps

from cloudify import ctx
from cloudify.context import NODE_INSTANCE
from cloudify.exceptions import NonRecoverableError

from .policies_output import PoliciesOutput

POLICIES = 'policies'
POLICY_FILTERS = 'policy_filters'
POLICIES_FILTERED = 'policies_filtered'

POLICY_ID = 'policy_id'
POLICY_BODY = 'policy_body'
POLICY_VERSION = "policyVersion"
POLICY_CONFIG = 'config'

POLICY_FILTER = 'policy_filter'
POLICY_FILTER_ID = 'policy_filter_id'

POLICY_PERSISTENT = 'policy_persistent'
DCAE_POLICY_TYPE = 'dcae.nodes.policy'
DCAE_POLICIES_TYPE = 'dcae.nodes.policies'

ACTION_GATHERED = "gathered"
ACTION_UPDATED = "updated"

CONFIG_ATTRIBUTES = "configAttributes"

class Policies(object):
    """static class for policy operations"""
    _updated_policies = {}
    _removed_policies = {}

    @staticmethod
    def _init():
        """init static members"""
        Policies._updated_policies = {}
        Policies._removed_policies = {}

    @staticmethod
    def _set_policies(policies):
        """store the :policies: in :runtime_properties[POLICIES]:"""
        if not policies:
            if POLICIES in ctx.instance.runtime_properties:
                del ctx.instance.runtime_properties[POLICIES]
            return

        ctx.instance.runtime_properties[POLICIES] = policies

    @staticmethod
    def _add_policy(policy_id, policy, policy_persistent, policies):
        """only add the latest version of policy to policies"""
        prev_policy = policies.get(policy_id)
        prev_policy_persistent = (prev_policy or {}).get(POLICY_PERSISTENT)

        if not prev_policy \
        or (policy_persistent and not prev_policy_persistent) \
        or (policy_persistent == prev_policy_persistent and policy.get(POLICY_BODY)):
            policy = deepcopy(policy)
            policy[POLICY_PERSISTENT] = policy_persistent
            policies[policy_id] = policy

    @staticmethod
    def _gather_policy(target, policies):
        """adds the policy from dcae.nodes.policy node to policies"""
        if DCAE_POLICY_TYPE not in target.node.type_hierarchy:
            return
        policy_id = target.node.properties.get(POLICY_ID)
        if not policy_id:
            return True
        policy = deepcopy(dict(target.node.properties))
        policy_body = target.instance.runtime_properties.get(POLICY_BODY)
        if policy_body:
            policy[POLICY_BODY] = policy_body

        Policies._add_policy(policy_id, policy, True, policies)
        return True

    @staticmethod
    def _fix_policy_filter(policy_filter):
        if CONFIG_ATTRIBUTES in policy_filter:
            config_attributes = policy_filter.get(CONFIG_ATTRIBUTES)
            if isinstance(config_attributes, dict):
                return
            try:
                config_attributes = json.loads(config_attributes)
                if config_attributes and isinstance(config_attributes, dict):
                    policy_filter[CONFIG_ATTRIBUTES] = config_attributes
                    return
            except (ValueError, TypeError):
                pass
            if config_attributes:
                ctx.logger.warn("unexpected %s: %s", CONFIG_ATTRIBUTES, config_attributes)
            del policy_filter[CONFIG_ATTRIBUTES]

    @staticmethod
    def _gather_policies(target, policies, policy_filters):
        """adds the policies and policy-filter from dcae.nodes.policies node to policies"""
        if DCAE_POLICIES_TYPE not in target.node.type_hierarchy:
            return

        property_policy_filter = target.node.properties.get(POLICY_FILTER)
        if property_policy_filter:
            policy_filter = deepcopy(dict(
                (k, v) for (k, v) in dict(property_policy_filter).items()
                if v or isinstance(v, (int, float))
            ))
            Policies._fix_policy_filter(policy_filter)

            if policy_filter:
                policy_filters[target.instance.id] = {
                    POLICY_FILTER_ID : target.instance.id,
                    POLICY_FILTER : policy_filter
                }

        filtered_policies = target.instance.runtime_properties.get(POLICIES_FILTERED)
        if not filtered_policies or not isinstance(filtered_policies, dict):
            return True
        for (policy_id, policy) in filtered_policies.items():
            Policies._add_policy(policy_id, policy, False, policies)
        return True

    @staticmethod
    def _get_policy_bodies_dict(policies):
        """returns a dict of policy_id -> policy_body if policy_body exists"""
        if not policies:
            return {}

        return {policy_id: policy.get(POLICY_BODY)
                for policy_id, policy in policies.items() if policy.get(POLICY_BODY)}

    @staticmethod
    def gather_policies_to_node():
        """
        decorate with @Policies.gather_policies_to_node() to
        gather the policies from dcae.nodes.policy nodes this node depends on.

        Places the policies into runtime_properties["policies"].

        Stores <scn>:policies data in consul-kv
        """
        def gather_policies_decorator(func):
            """the decorator"""
            if not func:
                return

            @wraps(func)
            def wrapper(*args, **kwargs):
                """gather and save the policies from dcae.nodes.policy nodes this node related to"""
                if ctx.type != NODE_INSTANCE:
                    raise NonRecoverableError("can only invoke gather_policies_to_node on node")

                policy_bodies = []
                try:
                    policies = {}
                    policy_filters = {}
                    for rel in ctx.instance.relationships:
                        _ = Policies._gather_policy(rel.target, policies) \
                         or Policies._gather_policies(rel.target, policies, policy_filters)

                    Policies._set_policies(policies)
                    if policy_filters:
                        ctx.instance.runtime_properties[POLICY_FILTERS] = policy_filters

                    policy_bodies = Policies._get_policy_bodies_dict(policies)
                except Exception as ex:
                    error = "Failed to set the policies {0}".format(str(ex))
                    ctx.logger.error("{0}: {1}".format(error, traceback.format_exc()))
                    raise NonRecoverableError(error)

                func_result = func(*args, **kwargs)

                if policy_bodies:
                    PoliciesOutput.store_policies(ACTION_GATHERED, policy_bodies)

                return func_result
            return wrapper
        return gather_policies_decorator

    @staticmethod
    def _update_policies(updated_policies, added_policies, removed_policies):
        """
        filter and update policies in runtime_properties
        and return the ordered filtered list of changed_policies
        """
        Policies._init()

        if not updated_policies and not removed_policies and not added_policies:
            ctx.logger.error("update_policies_on_ctx - no updated, added, or removed policies received")
            return

        updated_policies = updated_policies or []
        added_policies = added_policies or {}
        removed_policies = removed_policies or []

        ctx.logger.info("updated_policies: {0}, added_policies: {1}, removed_policies: {2}"
                        .format(json.dumps(updated_policies),
                                json.dumps(added_policies),
                                json.dumps(removed_policies)))

        policies = ctx.instance.runtime_properties.get(POLICIES, {})
        policy_filters = ctx.instance.runtime_properties.get(POLICY_FILTERS, {})

        for policy_id in removed_policies:
            removed_policy = policies.get(policy_id)
            if removed_policy and POLICY_BODY in removed_policy:
                Policies._removed_policies[policy_id] = deepcopy(removed_policy)
                if removed_policy.get(POLICY_PERSISTENT):
                    del policies[policy_id][POLICY_BODY]
                else:
                    del policies[policy_id]

        new_policies = dict((policy_id, policy)
                            for policy_filter_id in policy_filters
                            for (policy_id, policy) in added_policies.get(policy_filter_id, {})
                                                                     .get(POLICIES, {}).items())

        ctx.logger.info("new_policies: {0}".format(json.dumps(new_policies)))

        for policy_id in new_policies:
            policy = new_policies.get(policy_id)
            deployed_policy = policies.get(policy_id)
            if not deployed_policy:
                policies[policy_id] = policy
                Policies._updated_policies[policy_id] = policy
                continue
            updated_policies.append(policy)

        skipped = {"ignored": [], "unexpected": [], "same": [], "duplicate": []}
        for policy in updated_policies:
            policy_id = policy.get(POLICY_ID)
            if not policy_id or policy_id not in policies:
                skipped["ignored"].append(policy)
                continue

            if policy_id in Policies._updated_policies:
                skipped["duplicate"].append(policy)
                continue

            updated_policy_body = policy.get(POLICY_BODY, {})
            updated_policy_version = updated_policy_body.get(POLICY_VERSION)
            if not updated_policy_version or POLICY_CONFIG not in updated_policy_body:
                skipped["unexpected"].append(policy)
                continue

            deployed_policy = policies.get(policy_id)
            deployed_policy_version = deployed_policy.get(POLICY_BODY, {}).get(POLICY_VERSION)
            if updated_policy_version == deployed_policy_version:
                skipped["same"].append(policy)
                continue

            policies[policy_id][POLICY_BODY] = updated_policy_body
            Policies._updated_policies[policy_id] = policy

        if skipped["same"] or skipped["ignored"] or skipped["unexpected"] or skipped["duplicate"]:
            ctx.logger.info("skipped updates on policies: {0}".format(json.dumps(skipped)))

        if Policies._updated_policies or Policies._removed_policies:
            Policies._set_policies(policies)
            policy_bodies = Policies._get_policy_bodies_dict(policies)
            PoliciesOutput.store_policies(ACTION_UPDATED, policy_bodies)

    @staticmethod
    def update_policies_on_node():
        """
        decorate each policy_update operation with @Policies.update_policies_on_node() to
        filter out the updated_policies to only what applies to the current node instance,
        update runtime_properties["policies"]

        updates <scn>:policies data in consul-kv

        Passes through the filtered list of updated_policies that apply to the current node instance

        :updated_policies: contains the list of changed policy_bodies

        :removed_policies: contains the list of removed policy_bodies

        :policies: contains the list of current policy_bodies
        """
        def update_policies_decorator(func):
            """actual decorator"""
            if not func:
                return

            @wraps(func)
            def wrapper(updated_policies=None, added_policies=None, removed_policies=None, **kwargs):
                """update matching policies on the node"""
                if ctx.type != NODE_INSTANCE:
                    raise NonRecoverableError("can only invoke update_policies_on_node on node")

                try:
                    Policies._update_policies(updated_policies, added_policies, removed_policies)

                    updated_policies = Policies.get_policy_bodies(
                        selected_policies=Policies._updated_policies
                    )

                    removed_policies = Policies.get_policy_bodies(
                        selected_policies=Policies._removed_policies
                    )

                except Exception as ex:
                    error = "Failed to update the policies {0}".format(str(ex))
                    ctx.logger.error("{0}: {1}".format(error, traceback.format_exc()))
                    raise NonRecoverableError(error)

                if updated_policies or removed_policies:
                    return func(updated_policies,
                                removed_policies=removed_policies,
                                policies=Policies.get_policy_bodies(),
                                **kwargs)
            return wrapper
        return update_policies_decorator

    @staticmethod
    def cleanup_policies_on_node(func):
        """
        decorate each delete operation with @Policies.cleanup_policies_on_node to
        remove <scn>:policies data in consul-kv
        """
        if not func:
            return

        @wraps(func)
        def wrapper(**kwargs):
            """remove policies in consul-kv"""
            if ctx.type == NODE_INSTANCE:
                try:
                    PoliciesOutput.delete_policies()
                except Exception as ex:
                    error = "Failed to cleanup policies in consul-kv {0}".format(str(ex))
                    ctx.logger.error("{0}: {1}".format(error, traceback.format_exc()))

            return func(**kwargs)
        return wrapper

    @staticmethod
    def get_policy_bodies(selected_policies=None):
        """returns the list of policy_body objects if policy_body exists"""
        if isinstance(selected_policies, dict):
            return deepcopy([policy.get(POLICY_BODY)
                             for policy in selected_policies.itervalues() if policy.get(POLICY_BODY)])

        policies = ctx.instance.runtime_properties.get(POLICIES, {})
        return deepcopy([policy.get(POLICY_BODY)
                         for policy in policies.itervalues() if policy.get(POLICY_BODY)])



# ================================================================================
# Copyright (c) 2017-2019 AT&T Intellectual Property. All rights reserved.
# ================================================================================
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============LICENSE_END=========================================================
#
# ECOMP is a trademark and service mark of AT&T Intellectual Property.

"""setup.py is used for package build and distribution"""

from setuptools import setup, find_packages

setup(
    name='onap-dcae-dcaepolicy-lib',
    description='lib of policy decorators to be used by cloudify plugins of dcae controller',
    version="2.5.1",
    author='Alex Shatov',
    author_email="alexs@att.com",
    license='Apache 2',
    packages=find_packages(),
    install_requires=[
    ],
    keywords='onap policy dcae controller cloudify plugin',
    classifiers=[
        'Intended Audience :: Telecommunications Industry',
        'Programming Language :: Python :: 2.7'
    ]
)



# ================================================================================
# Copyright (c) 2019-2021 AT&T Intellectual Property. All rights reserved.
# ================================================================================
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============LICENSE_END=========================================================
#
# ECOMP is a trademark and service mark of AT&T Intellectual Property.

"""client to talk to consul on standard port 8500"""

import base64
import json
try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote
import uuid
from datetime import datetime

import requests
from cloudify import ctx


class PoliciesOutput(object):
    """static class for store-delete policies in consul kv"""
    # it is safe to assume that consul agent is at consul:8500
    # define consul alis in /etc/hosts on cloudify manager vm
    # $ cat /etc/hosts
    # 127.0.0.1   localhost localhost.localdomain localhost4 localhost4.localdomain4 consul
    CONSUL_TRANSACTION_URL = "http://consul:8500/v1/txn"

    POLICIES_EVENT = 'policies_event'

    POLICIES_FOLDER_MASK = "{0}:policies/{1}"
    MAX_OPS_PER_TXN = 64

    OPERATION_SET = "set"
    OPERATION_DELETE = "delete"
    OPERATION_DELETE_FOLDER = "delete-tree"
    SERVICE_COMPONENT_NAME = "service_component_name"


    @staticmethod
    def _gen_txn_operation(verb, service_component_name, key=None, value=None):
        """returns the properly formatted operation to be used inside transaction"""
        key = PoliciesOutput.POLICIES_FOLDER_MASK.format(
            service_component_name, quote(key or "")
        )
        if value:
            return {"KV": {"Verb": verb, "Key": key, "Value": base64.b64encode(value.encode("utf8")).decode("utf8")}}
        return {"KV": {"Verb": verb, "Key": key}}


    @staticmethod
    def _run_transaction(operation_name, txn):
        """run a single transaction of several operations at consul /txn"""
        if not txn:
            return None

        response = None
        try:
            response = requests.put(PoliciesOutput.CONSUL_TRANSACTION_URL, json=txn)
        except requests.exceptions.RequestException as ex:
            ctx.logger.error(
                "RequestException - failed to {0} at {1}: {2} on txn={3}"
                .format(operation_name, PoliciesOutput.CONSUL_TRANSACTION_URL,
                        str(ex), json.dumps(txn)))
            return None

        if response.status_code != requests.codes.ok:
            ctx.logger.error(
                "failed {0} for {1} {2}: text={3} txn={4}"
                .format(response.status_code, operation_name,
                        PoliciesOutput.CONSUL_TRANSACTION_URL, response.text, json.dumps(txn)))
            return None
        ctx.logger.info(
            "response {0} for {1} {2}: text={3} txn={4}"
            .format(response.status_code, operation_name,
                    PoliciesOutput.CONSUL_TRANSACTION_URL, response.text, json.dumps(txn)))
        return True


    @staticmethod
    def store_policies(action, policy_bodies):
        """put the policy_bodies for service_component_name into consul-kv"""
        service_component_name = ctx.instance.runtime_properties.get(
            PoliciesOutput.SERVICE_COMPONENT_NAME
        )
        if not service_component_name:
            ctx.logger.warn("failed to find service_component_name to store_policies in consul-kv")
            return False

        event = {
            "action": action,
            "timestamp": (datetime.utcnow().isoformat()[:-3] + 'Z'),
            "update_id": str(uuid.uuid4()),
            "policies_count": len(policy_bodies)
        }
        ctx.instance.runtime_properties[PoliciesOutput.POLICIES_EVENT] = event

        store_policies = [
            PoliciesOutput._gen_txn_operation(PoliciesOutput.OPERATION_SET, service_component_name,
                                              "items/" + policy_id, json.dumps(policy_body))
            for policy_id, policy_body in policy_bodies.items()
        ]
        txn = [
            PoliciesOutput._gen_txn_operation(
                PoliciesOutput.OPERATION_DELETE_FOLDER, service_component_name),
            PoliciesOutput._gen_txn_operation(
                PoliciesOutput.OPERATION_SET, service_component_name, "event", json.dumps(event))
        ]
        idx_step = PoliciesOutput.MAX_OPS_PER_TXN - len(txn)
        for idx in range(0, len(store_policies), idx_step):
            txn += store_policies[idx : idx + idx_step]
            if not PoliciesOutput._run_transaction("store_policies", txn):
                return False
            txn = []

        PoliciesOutput._run_transaction("store_policies", txn)
        return True


    @staticmethod
    def delete_policies():
        """delete policies for service_component_name in consul-kv"""
        if PoliciesOutput.POLICIES_EVENT not in ctx.instance.runtime_properties:
            return

        service_component_name = ctx.instance.runtime_properties.get(
            PoliciesOutput.SERVICE_COMPONENT_NAME
        )
        if not service_component_name:
            ctx.logger.warn("failed to find service_component_name to delete_policies in consul-kv")
            return

        delete_policies = [
            PoliciesOutput._gen_txn_operation(
                PoliciesOutput.OPERATION_DELETE_FOLDER, service_component_name
            )
        ]
        PoliciesOutput._run_transaction("delete_policies", delete_policies)
